"""CarbonTracker Data Assimilation Shell (CTDAS) Copyright (C) 2017 Wouter Peters. 
Users are recommended to contact the developers (wouter.peters@wur.nl) to receivstope
updates of the code. See also: http://www.carbontracker.eu. 

This program is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software Foundation, 
version 3. This program is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this 
program. If not, see <http://www.gnu.org/licenses/>."""
#!/usr/bin/env python
# expand_fluxes.py
"""
Author: Joram Hooghiem 
Revision History:
File created on 05 May 2022.
based on original version by Wouter Peters
"""

import sys
import subprocess
import os
sys.path.append('../../') 
rootdir = os.getcwd().split('da/')[0] 
analysisdir = os.path.join(rootdir, 'da/analysis/cteco2/')


from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
import shutil

import logging
import numpy as np
from da.tools.general import date2num, num2date, create_dirs,advance_time,datevector
import da.tools.io4 as io
import netCDF4 as cdf

from   da.analysis.cteco2.tools_regions import globarea, state_to_grid
import da.analysis.cteco2.tools_transcom as tc 
import da.analysis.cteco2.tools_country as ct
import da.analysis.cteco2.tools_time as timetools
from da.ctdata.fluxload import load_wrapper
refyear=datetime(2000,1,1)
def save_1x1_data(dacycle, statevector):
    """
        Function creates a NetCDF file with output on 1x1 degree grid. It uses the flux data written by the 
        :class:`~da.baseclasses.obsoperator.ObsOperator.py`, and multiplies these with the mapped parameters and
        variance (not covariance!) from the :class:`~da.baseclasses.statevector.StateVector`.
        
           :param dacycle: a :class:`~da.tools.initexit.CycleControl` object
           :param statevector: a :class:`~da.baseclasses.statevector.StateVector`
           :rtype: None
    """
    for date in datevector(dacycle['time.start'],dacycle['time.end'],'day')[:-1]:
        write_flux1x1(dacycle,statevector,date)
    return
def save_agg_data(dacycle,region_aggregate):
    for date in datevector(dacycle['time.start'],dacycle['time.end'],'day')[:-1]:
        write_agg(dacycle,date,region_aggregate=region_aggregate)
    return
###
def write_flux1x1(dacycle,statevector,date,alltimes=False,imposed=True,basename="flux1x1.",fmt='%Y-%m-%d',key=None):

    
    if alltimes:
        datasets=load_data(dacycle,statevector,date,mean_only=True,timeavg=False,imposed=imposed,key=key)
        for key,data in datasets.items():
            # shape is either (days, dpd,180,360) 
            # or days 180
            nr=np.shape(data)[2]
        # assumie 3 hourly for now....
        # hours per point 
        if nr == 1:
            ndates=[date2num(date + timedelta(0.5),refyear=refyear)]
        else:
            hpp=24/nr
            ndates=[date2num(date + timedelta(hours=hpp*i),refyear=refyear) for i in range(0,nr)]
        members=1
    else:
        ndates=[date2num(date+timedelta(0.5),refyear=refyear) ]
        members=statevector.nmembers
        datasets=load_data(dacycle,statevector,date)

    dirname = create_dirs(os.path.join(dacycle['dir.analysis'], 'data_flux1x1_daily'))
    logging.info("Opening %s for writing prior and posterior fluxes" )     
    ncf = io.CT_CDF(os.path.join(dirname, '%s%s.nc' % (basename,date.strftime(fmt))), 'write')
    setattr(ncf, 'Title', 'CarbonTracker fluxes')
    setattr(ncf, 'node_offset', 1)
    dimgrid = ncf.add_latlon_dim()
    dimensemble = ncf.add_dim('members', members)
    dimdate = ncf.add_date_dim()

    save_data(ncf,'cell_area',globarea().tolist(),dimgrid) 
    next=0
    for i,ndate in enumerate(ndates):
        save_data(ncf,'date',ndate,dimdate,next) 
        for key,data in datasets.items():
            # logging.info(np.shape(data))
            if '_ensemble' in key:   
                save_data(ncf,key,data[:,0,i].tolist(),dimdate + dimensemble + dimgrid ,next) 
            else:
                save_data(ncf,key,data[0,i].tolist(),dimdate +  dimgrid ,next) 

        next = ncf.inq_unlimlen()[0]
    ncf.close()

    # logging.info("Gridded weekly average fluxes now written")
    return

def write_agg(dacycle,date, region_aggregate='olson'):
    """
        Function creates a NetCDF file with output on TransCom regions. It uses the flux input from the 
        function `save_weekly_avg_1x1_data` to create fluxes of length `nparameters`, which are then projected
        onto TC regions using the internal methods from :class:`~da.baseclasses.statevector.StateVector`.
        
           :param dacycle: a :class:`~da.tools.initexit.CycleControl` object
           :param StateVector: a :class:`~da.baseclasses.statevector.StateVector`
           :rtype: None

        This function only read the prior fluxes from the flux_1x1.nc files created before, because we want to convolve 
        these with the parameters in the statevector. This creates posterior fluxes, and the posterior covariance for the complete
        statevector in units of mol/box/s which we then turn into TC fluxes and covariances.
    """

    dirname = create_dirs(os.path.join(dacycle['dir.analysis'], 'data_%s_daily' % (region_aggregate)))
    
    # ndate=date2num(proc_data_b) - date2num(datetime(2000, 1, 1)) + (proc_data_e-proc_data_b).days / 2.0
    # logging.info('Processing output emissions from %s to %s' %(proc_data_b.strftime('%Y-%m-%d %H:%M'),proc_data_e.strftime('%Y-%m-%d %H:%M'))) 

    infile = os.path.join(dacycle['dir.analysis']             , 'data_flux1x1_daily' , 'flux1x1.%s.nc' % date.strftime('%Y-%m-%d'))
    if not os.path.exists(infile):
        logging.error("Needed input file (%s) does not exist yet, please create file first, returning..." % infile)
        return None


    ncf = io.CT_CDF(os.path.join(dirname, '%s_fluxes.%s.nc' % (region_aggregate,date.strftime('%Y-%m-%d'))), 'write')

    dimdate = ncf.add_date_dim()
    dimidateformat = ncf.add_date_dim_format()
    dimgrid = ncf.add_latlon_dim()  # for mask
    regionmask,dimname,dimregs,xform=get_region_data(ncf,region_aggregate)

    savedict = ncf.standard_var('unknown')
    savedict['name'] = 'regionmask'
    savedict['comment'] = 'numerical mask used to aggregate 1x1 flux fields, each integer 0,...,N is one region aggregated'
    savedict['values'] = regionmask.tolist()
    savedict['units'] = '-'
    savedict['dims'] = dimgrid
    savedict['count'] = 0
    ncf.add_data(savedict)

    # Get input data from 1x1 degree flux files
    ncf_in = io.ct_read(infile, 'read')
    dates = ncf_in.variables['date'][:]
    # First add the date for this cycle to the file, this grows the unlimited dimension
    for index,date in enumerate(dates):
        save_data(ncf,'date',dates[index], dimdate,count=index)

    # Now convert other variables that were inside the statevector file

        # get area
        area = globarea()

        vardict = ncf_in.variables
        for vname, vprop in vardict.items():
            # logging.debug('processing %s' % vname)
            if vname in ['latitude','longitude', 'date', 'idate', 'std']: continue
            elif 'ensemble' in vname:

                data = ncf_in.get_variable(vname)[index]
         
                dimensemble = ncf.add_dim('members', data.shape[0])

                regiondata = []
                for member in data:
                    aggdata = state_to_grid(member * area, regionmask, reverse=True, mapname=region_aggregate)
                    regiondata.append(aggdata)

                regiondata = np.array(regiondata)
                try:
                    regioncov = regiondata.transpose().dot(regiondata) / (data.shape[0] - 1) 
                except:
                    regioncov = np.dot(regiondata.transpose(), regiondata) / (data.shape[0] - 1) # Huygens fix 

                if xform:
                    regioncov = tc.ExtendedTCRegions(regioncov,cov=True)
            
                savedict = ncf.standard_var(varname=vname)
                savedict['name'] = vname.replace('ensemble','covariance') 
                savedict['units'] = '[mol/region/s]^2'
                savedict['dims'] = dimdate + dimregs + dimregs
                savedict['count'] = index
                savedict['values'] = regioncov
                ncf.add_data(savedict)

                unit='mol/region/s'
                dims=dimdate + dimensemble + dimregs
            
            elif 'flux' in vname:

                data = ncf_in.get_variable(vname)[index]
                regiondata = state_to_grid(data * area, regionmask, reverse=True, mapname=region_aggregate)

                dims = dimdate + dimregs
                unit = 'mol/region/s'

            else:

                data = ncf_in.get_variable(vname)[:]
                regiondata = state_to_grid(data, regionmask, reverse=True, mapname=region_aggregate)
                dims = dimdate + dimregs
                unit = ''

            if xform:
                regiondata = tc.ExtendedTCRegions(regiondata)
            savedict = ncf.standard_var(varname=vname)
            if not unit =='':
                savedict['units'] = unit 
            savedict['name'] = vname
            savedict['dims'] = dims 
            savedict['count'] = index
            savedict['values'] = regiondata
            ncf.add_data(savedict)

        ncf_in.close()
        ncf.close()
    # logging.info("%s aggregated %s average fluxes now written" % (dimname,interval))

    return 




def save_data(ncf,varname,data,dims,count=None):
    logging.info("Writing variable %s to file" % varname )
    savedict = ncf.standard_var(varname=varname)
    savedict['values'] = data
    savedict['dims'] = dims
    if not count==None:
        savedict['count'] = count
    ncf.add_data(savedict)
    return

def get_input(dacycle,date):
    """Docstring for get_input_file.
    opens an ncfile using the io4 ct_read
    filename will be constructed from startdate en enddate TYPE? 

    Parameters
    ----------
    species : string of the species name, which
    startdate :
    enddate :
    returns : carbontracker netcdf dataset object
    """
  
 
    fluxes_imposed={}
    for species in dacycle.dasystem['ct.params.species'].split(' '):
        filename = os.path.join(dacycle['dir.output'], 'flux1x1_%s_%s_%s.nc' % (species,date.strftime('%Y%m%d%H'),(date+timedelta(1)).strftime('%Y%m%d%H')))
        logging.info("fetching data from %s" % filename)
        ncf = io.ct_read(filename, 'read')
        for key in ncf.variables.keys():
            if key in ['lon','lat','lev']: continue
            # logging.info("fetching data for variable %s" %key)
            data=ncf.variables[key][:]
            if len(np.shape(data))>2:
                logging.info('Currently only surface emissions are implemented') 
                logging.info('3D emissions require the a and b hybrid sigma coefs and surface pressure from obsoperator') 
                logging.info('Here we reduce the flux in mol/s to a 2d lat lon flux in mol/m2/s') 
                data=np.sum(data,axis=0)/globarea()
            fluxes_imposed.update({fetchkey(species,key.split('_')[2]):data})
        ncf.close()
    return fluxes_imposed 


def fetchkey(species,emission):
    if emission in ['ff']:
        emissionfixed='fossil'
    elif emission in ['fire']:
        emissionfixed='fires'
    elif emission in ['biodiu']:
        emissionfixed='bio'
    else:
        emissionfixed=emission
    key='%s_%s_flux_imp' %(species,emissionfixed) 
    return key


def get_region_data(ncf,region_aggregate):
#
#   Select regions to aggregate to
# 
    if region_aggregate == "olson":
        regionmask = tc.olson240mask
        dimname = 'olson'
        dimregs = ncf.add_dim(dimname, regionmask.max())

        regionnames = []
        for i in range(11):
            for j in range(19):
                regionnames.append("%s_%s" % (tc.transnams[i], tc.olsonnams[j],))
        regionnames.extend(tc.oifnams)
        xform = False

        for i, name in enumerate(regionnames):
            lab = 'Aggregate_Region_%03d' % (i + 1,)
            setattr(ncf, lab, name)

    elif region_aggregate == "olson_extended":
        regionmask = tc.olson_ext_mask
        dimname = 'olson_ext'
        dimregs = ncf.add_dim(dimname, regionmask.max())
        xform = False

        for i, name in enumerate(tc.olsonextnams):
            lab = 'Aggreate_Region_%03d'%(i+1)
            setattr(ncf, lab, name)

    elif region_aggregate == "transcom":
        regionmask = tc.transcommask
        dimname = 'tc'
        dimregs = ncf.add_region_dim(type='tc')
        xform = False

    elif region_aggregate == "transcom_extended":
        regionmask = tc.transcommask
        dimname = 'tc_ext'
        dimregs = ncf.add_region_dim(type='tc_ext')
        xform = True
    
    elif region_aggregate == "amazon":
        regfile = cdf.Dataset(os.path.join(analysisdir,'amazon_mask.nc'))
        regionmask = regfile.variables['regionmask'][:]
        regfile.close()
        dimname = 'amazon'
        dimregs = ncf.add_dim(dimname, regionmask.max())
        xform = False

    elif region_aggregate == "country":

        xform = False
        countrydict = ct.get_countrydict()
        selected = ['Russia', 'Canada', 'China', 'United States', 'EU27', 'Brazil', 'Australia', 'India'] #,'G8','UNFCCC_annex1','UNFCCC_annex2']
        regionmask = np.zeros((180, 360,), 'float')

        for i, name in enumerate(selected):
            lab = 'Country_%03d' % (i + 1,)
            setattr(ncf, lab, name)

            if name == 'EU27':
                namelist = ct.EU27
            elif name == 'EU25':
                namelist = ct.EU25
            elif name == 'G8':
                namelist = ct.G8
            elif name == 'UNFCCC_annex1':
                namelist = ct.annex1
            elif name == 'UNFCCC_annex2':
                namelist = ct.annex2
            else:
                namelist = [name]

            for countryname in namelist:
                try:
                    country = countrydict[countryname]
                    regionmask.put(country.gridnr, i + 1)
                except:
                    continue

        dimname = 'country'
        dimregs = ncf.add_dim(dimname, regionmask.max())

    return regionmask,dimname,dimregs,xform

def time_avg(dacycle,avg='transcom',date=None):
    """ Function to create a set of averaged files in a folder, needed to make longer term means """
    
    if avg not in ['transcom','transcom_extended','olson','olson_extended','country','flux1x1']:
        raise IOError('Choice of averaging invalid')

    if date==None:
        date=dacycle['time.start']

    analysisdir = dacycle['dir.analysis']

    if not os.path.exists(analysisdir):
        raise IOError('analysis dir requested (%s) does not exist, exiting...'%analysisdir)

    monthly_avg(dacycle,avg,date)
    if date.month==12:
        yearly_avg(dacycle,avg,date)
    # longterm_avg(dacycle,avg)

    return
def new_month(dacycle):
    """ check whether we just entered a new month"""

    this_month = dacycle['time.start'].month
    prev_month = (dacycle['time.start']-dacycle['cyclelength']).month

    return (this_month != prev_month)

def new_year(dacycle):
    """ check whether we just entered a new year"""

    this_year = dacycle['time.start'].year
    prev_year = (dacycle['time.start']-dacycle['cyclelength']).year

    return (this_year != prev_year)

def daily_avg(dacycle,avg):
    """ Function to create a set of daily files in a folder, needed to make longer term means """
    
    if avg not in ['transcom','transcom_extended','olson','olson_extended','country','flux1x1']:
        raise IOError('Choice of averaging invalid')

    analysisdir = dacycle['dir.analysis']
    weekdir = os.path.join(analysisdir , 'data_%s_weekly'%avg)
    daydir = os.path.join(analysisdir , 'data_%s_daily'%avg)

    if not os.path.exists(daydir):
        print("Creating new output directory " + daydir)
        os.makedirs(daydir)

    files  = os.listdir(weekdir)
    files = [f for f in files if '-' in f and f.endswith('.nc')]

    fileinfo = {}
    for filename in files:
        date=datetime.strptime(filename.split('.')[-2],'%Y-%m-%d')
        fileinfo[filename] = date
    
    dt = dacycle['cyclelength']

    for k,v in fileinfo.items():
        cycle_file = os.path.join(weekdir,k)
        for i in range(abs(dt.days)):
            daily_file = os.path.join(daydir,'%s_fluxes.%s.nc'%(avg,(v+timedelta(days=i)).strftime('%Y-%m-%d')))
            if not os.path.lexists(daily_file):
                os.symlink(cycle_file,daily_file)
                #print daily_file,cycle_file

def monthly_avg(dacycle,avg,sd):
    """ Function to average a set of files in a folder from daily to monthly means """
    
    if avg not in ['transcom','transcom_extended','olson','olson_extended','country','flux1x1']:
        raise IOError('Choice of averaging invalid')

    analysisdir = dacycle['dir.analysis']

    daydir = os.path.join(analysisdir , 'data_%s_daily'%avg)
    monthdir = os.path.join(analysisdir,'data_%s_monthly'%avg)

    if not os.path.exists(monthdir):
        print("Creating new output directory " + monthdir)
        os.makedirs(monthdir)


    files  = os.listdir(daydir)  # get daily files
    files = [f for f in files if '-' in f and f.endswith('.nc')]

    if len(files) < 28:
        print('No month is yet complete, skipping monthly average')
        return

    fileinfo = {}
    for filename in files:  # parse date from each of them
        date=datetime.strptime(filename.split('.')[-2],'%Y-%m-%d')
        fileinfo[filename] = date

    years = [d.year for d in list(fileinfo.values())]   # get actual years
    months = set([d.month for d in list(fileinfo.values())])  # get actual months
   
    # sd = datein 
    # sd = datetime(min(year),1,1)
    # ed = datetime(max(year)+1,1,1)

    # while sd < ed: 

    nd = sd + relativedelta(months=+1)

    ndays_in_month = (nd-sd).days
    
    avg_files = [os.path.join(daydir,k) for k,v in fileinfo.items() if v < nd and v >= sd]
    
    if len(avg_files) != ndays_in_month: # only once month complete 
        #print 'New month (%02d) is not yet complete, skipping monthly average'%(sd.month)
        pass
    else:
        targetfile = os.path.join(monthdir,'%s_fluxes.%s.nc'%(avg,sd.strftime('%Y-%m')))
        if not os.path.exists(targetfile):
            print("New month (%02d) is complete, I have %d days for the next file"%(sd.month,ndays_in_month))
            command = ['ncra','-O']+ avg_files + [targetfile]
            status = subprocess.check_call(command)
        else:
            pass

    # sd = nd

def yearly_avg(dacycle,avg,datein):
    """ Function to average a set of files in a folder from monthly to yearly means """

    if avg not in ['transcom','transcom_extended','olson','olson_extended','country','flux1x1']:
        raise IOError('Choice of averaging invalid')

    analysisdir = dacycle['dir.analysis']
    monthdir = os.path.join(analysisdir , 'data_%s_daily'%avg )
    yeardir = os.path.join(analysisdir,'data_%s_yearly'%avg)

    if not os.path.exists(yeardir):
        print("Creating new output directory " + yeardir)
        os.makedirs(yeardir)

    files  = os.listdir(monthdir)  # get monthly files
    files = [f for f in files if '-' in f and f.endswith('.nc')]

    if not files:
        print("No full year finished yet, skipping yearly average...")
        return

    fileinfo = {}
    for filename in files:
        date=datetime.strptime(filename.split('.')[-2],'%Y-%m-%d')
        fileinfo[filename] = date

    years = set([d.year for d in list(fileinfo.values())])

    sd = datetime(datein.year,1,1)
    # ed = datetime(max(year)+1,1,1)

    # while sd < ed: 

    nd = sd + relativedelta(years=+1)
    
    avg_files = [os.path.join(monthdir,k) for k,v in fileinfo.items() if v < nd and v >= sd]
    
    if not len(avg_files) == (nd-sd).days : 
        print("Year %04d not finished yet, skipping yearly average..."%sd.year)
    else:
        targetfile = os.path.join(yeardir,'%s_fluxes.%s.nc'%(avg,sd.strftime('%Y')))
    
        if not os.path.exists(targetfile):
            print("Year %04d is complete, I have 12 months for the next file"%sd.year)
            command = ['ncra','-O']+ avg_files + [targetfile]
            status = subprocess.check_call(command)


def longterm_avg(dacycle,avg):
    """ Function to average a set of files in a folder from monthly to yearly means """

    if avg not in ['transcom','transcom_extended','olson','olson_extended','country','flux1x1']:
        raise IOError('Choice of averaging invalid')

    analysisdir = dacycle['dir.analysis']

    yeardir = os.path.join(analysisdir , 'data_%s_yearly'%avg )
    longtermdir = os.path.join(analysisdir,'data_%s_longterm'%avg)

    if not os.path.exists(longtermdir):
        print("Creating new output directory " + longtermdir)
        os.makedirs(longtermdir)

    files  = os.listdir(yeardir)
    files = [f for f in files if f.endswith('.nc')]
    if not files:
        print("No full year finished yet, skipping longterm average...")
        return

    dates = []
    for filename in files:
        date=datetime.strptime(filename.split('.')[-2],'%Y')
        dates.append( date )

    avg_files = [os.path.join(yeardir,k) for k in files]
    targetfile=os.path.join(longtermdir,'%s_fluxes.%04d-%04d.nc'%(avg,min(dates).year, max(dates).year)) 
    if not os.path.exists(targetfile):
        if len(avg_files) > 0 : 
            command = ['ncra','-O']+ avg_files + [targetfile]
            status = subprocess.check_call(command)
    return

def load_data(dacycle,statevector,date,mean_only=False,timeavg=True,imposed=True,key=None):
    # get posterior and prior fluxes
    fluxes=statevector.state_to_grid(date=date,qual='prior',mean_only=mean_only,timeavg=timeavg,key=key) | statevector.state_to_grid(date=date,qual='opt',mean_only=mean_only,timeavg=timeavg,key=key)
    if imposed:
        fluxes.update(get_input(dacycle,date))
    for key in dacycle.dasystem['fluxes.expand.extra'].split(' '):
        if key=='': continue
        logging.info(key) 
        a,n=load_wrapper(key,[date],False,timeavg=timeavg)
        species,emission=n.split('_')
        fluxes.update({fetchkey(species,emission):a})
    # for key in imposed:
    #     data,ctname=ctdata.fetch_data(imposed,date)
    #     if len(np.shape(data[0]))>3 and daily_avg:
    #         data=np.mean(data,axis=0)
    #     fluxes.update({ctname : data} )
    fix_dims(fluxes)
    return fluxes

def fix_dims(fluxes):
    for key,data in fluxes.items():
        # logging.info(np.shape(data))
        if "imp" in key and len(np.shape(data))==3:
            fluxes[key]=data[np.newaxis]
        elif "imp" in key and len(np.shape(data))==2:
            fluxes[key]=data[np.newaxis,np.newaxis]
    return  

    
