"""CarbonTracker Data Assimilation Shell (CTDAS) Copyright (C) 2017 Wouter Peters.
Users are recommended to contact the developers (wouter.peters@wur.nl) to receive
updates of the code. See also: http://www.carbontracker.eu.

This program is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software Foundation,
version 3. This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this
program. If not, see <http://www.gnu.org/licenses/>."""
#!/usr/bin/env python
import os
import sys
sys.path.append(os.getcwd())
import numpy as np
import da.tools.io4 as io
from da.tools.general import datevector,advance_time
from da.statevectors.statevector_baseclass import StateVector, EnsembleMember
from da.statevectors.statevector_collections import statevector_collection
from da.tools.StateParameter import StateParameter 
import logging
identifier = 'CarbonTracker Statistical fit'
version = '0.0'

################### Start Class LWStateVector ###################
class StateVectorLW(StateVector):

    """
    This is a StateVector object for CarbonTracker. It is created for use with the SIF inversion approach,
    and contains statistical parameters and SIF sensitivity parameters per ecoregion to be assembled.
    It has a private method to make new ensemble members.
    """

    def setup(self, dacycle):
        """
        setup the object by specifying the dimensions.
        There are two major requirements for each statvector that you want to build:
            (1) is that the statevector can map itself onto a regular grid
            (2) is that the statevector can map itself (mean+covariance) onto TransCom regions
        An example is given below.
        """

        self.dacycle = dacycle

        logging.info('Calculating fluxes from %s to %s' % (self.dacycle['time.begin'],self.dacycle['time.finish']))

        self.isOptimized=dacycle['da.optimized']
        self.LW=dacycle.dasystem['ct.enkf.type']=='LW'
        logging.info("LW run is set to %s" %self.LW)
        # Temporal settings
        self.nlag       = int(dacycle['time.nlag'])
        if not self.LW:
            self.timecycle  = int(dacycle['time.cycle'])
            logging.info("timecyle %s" % self.timecycle)
        self.datevec    = np.array(datevector(self.dacycle['time.begin'],self.dacycle['time.finish'],'day'))[:-1]

        # Inversion setup
        self.nmembers   = int(dacycle['da.optimizer.nmembers'])                 # number of ensemble members
        self.statevectorname = dacycle.dasystem['ct.statevector.name']
        self.nobs              = 0
        self.obs_to_assimilate = ()                                             # empty containter to hold observations to assimilate later on
    

        state_settings={  "startdate" : dacycle['time.begin'],
                          "enddate"   : dacycle['time.finish'],
                          "nmembers"  : self.nmembers,
                          "nlag"      : self.nlag,
                          "lw"        : self.LW}
        # These list objects hold the data for each time step of lag in the system. Note that the ensembles for each time step consist
        # of lists of EnsembleMember objects, we define member 0 as the mean of the distribution and n=1,...,nmembers as the spread.

        # This specifies the file to read with the gridded mask at 1x1 degrees. Each gridbox holds a number that specifies the ecoregion it belongs to.
        # From this map, a dictionary is created that allows a reverse look-up so that we can map parameters to a grid.
        # logging.debug('There are %s non-empty ecoregions: %s' %(self.nregions, self.regionnumbers))
        # logging.debug('A dictionary to map grids to states and vice versa was created')
        # logging.debug('Average latitude per (nonempty) ecoregion was calculated.')
        # logging.info("Updating with dacycle.dasystem['nparameters']=%d to be used by the optimizer" %self.nparams )
        self.parameters_init = statevector_collection[self.statevectorname] 
        # logging.debug("%s" % self.nee_U)
        self.parameters={}
        for key,args in self.parameters_init.items():
            args.update(state_settings)
            self.parameters[key]=StateParameter(**args)
            logging.debug("setup a stateparameter '%s'" % (key))
        self.dacycle.dasystem['ct.params.names']=' '.join(self.parameters.keys())
        self.dacycle.dasystem['my.statevector']='with_statevector'
        if self.LW:
            self.dacycle.dasystem['ct.params.frequency']='daily'
        else:
            self.dacycle.dasystem['ct.params.frequency']='fixed'

        # by default select all
        self._select()
        return

    def write_to_file(self,filename=None):
        """ do whatever super does
        """
        if self.LW and self.dacycle['time.restart']:
            logging.info("not at the start of the run, wil not overwrite file")
            return
        if not len(self.loaded_keys)==len(self.parameters_init.keys()):
            #by default selects all when writing the statevector to file
            self._select()
            self._stateparameter2members()
        # file, dim of nparams ,nmembers, dimlag
        qual='prior'
        if self.isOptimized:
            qual='opt'
        if filename is None:
            filename=self._get_filename()
        
        f,dimparams, dimmembers, dimlag ,filename=super().write_to_file(filename,qual,return_file=True)
        f.close()
        logging.info('successfully wrote the %s state Vector to file (%s) ' % (qual,filename))
        return
    
    def read_from_file(self, filename=None, qual='opt',mem2state=True,date=None):
        """ read initial data for the statevector from given file 
        lag: integer, currently only needed for backwards compatibility with pipelines 
        qual: string quality of the state: 'opt' or 'prior'
        """

        if self.LW and not self.isOptimized:
            qual='prior'
        if filename==None:
            filename=self._get_filename(date=date,opt=qual,shift=True) # shift is used only in case of laged kalman filter and will read from previous cycle
        if not len(self.loaded_keys)==len(self.parameters_init.keys()):
            #by default selects all when reading the statevector from file
            self._select()

        logging.info("reading %s state from %s" %(qual,filename))
        super().read_from_file(filename,qual) 
        logging.info('successfully read the statevector from file (%s) ' % filename)
        if mem2state:
            self._members2stateparameter()
            self._setdomaininfo()
        logging.info('successfully assigned members back to state ')
        return

    def _members2stateparameter(self):
        for key in self.loaded_keys:
            for lag in range(0,self.nlag):
                self.parameters[key].members_to_state(self.ensemble_members[lag],*self.StateParameter_indices[key],lag)
            logging.info("Returning statevector elements for %s to parameter object" % key)
        return

    def _stateparameter2members(self):
        for key in self.loaded_keys:
            for lag in range(0,self.nlag):
                self.parameters[key].state_to_members(self.ensemble_members[lag],*self.StateParameter_indices[key],lag)
            logging.info("Inserting statevector elements for %s into the statevector for lag %s" % (key,lag))
        return

    def _setdomaininfo(self):
        if not self.LW:
            return
        for key in self.loaded_keys:
            self.parameters[key].domain_info(self.domain_info,*self.StateParameter_indices[key])
            logging.info("Inserting statevector elements for %s into the statevector" % key)
        return

    def state_to_grid(self, fluxvector=None,date=None, lag=1,monthcounter=None,filename=None,qual=None,mean_only=False,timeavg=True,key=None):
        """
            transforms the statevector information (mean + covariance) to a 1x1 degree grid.


            :param: fluxvector: a vector of length (nparams,) that holds the fluxes associated with each parameter in the StateVector
            :param: lag: the lag at which to evaluate the StateVector
            :rtype: a tuple of two arrays (gridmean,gridvariance) with dimensions (nfitparams,180,360)
            counter will fetch the fluxvalues for month=counter

            if the attribute `fluxvector` is not passed, the function will return the mean parameter value and its variance on a 1x1 map.

            ..note:: although we can return the variance information for each gridbox, the covariance information contained in the original ensemble is lost when mapping to 1x1 degree!
        """

        self.read_from_file(qual=qual,date=date)
        lag=self._date2lag(qual=qual,date=date)
        # currentdate=self.datevec[0]+relativedelta(months=monthcounter)
        
        # mask=[(currentdate.month==t.month)&(currentdate.year==t.year) for t in self.datevec]
        
        # dates= self.datevec[mask]
        results ={}
        
        for k,pobject in self.parameters.items():
            if not key is None:
                if k!=key: continue
            logging.debug('using %s to calculated %s %s' % (k,qual,pobject.fluxname))
            results.update(pobject.prior2post(np.array([date]),qual,mean_only=mean_only,timeavg=timeavg,lag=lag))
            for coupled_key in pobject.coupled:
                if coupled_key.split('_')[0] in self.dacycle.dasystem['ct.params.species'].split(' '): 
                    results.update(pobject.prior2post(np.array([date]),qual,mean_only=mean_only,timeavg=timeavg,lag=lag,altkey=coupled_key))

        # Get the mean flux:
        results_mean = { key.replace('_ensemble','') : results[key][0] for key in results.keys() if '_ensemble' in key  }
        if mean_only:
            return  results_mean
        else:
            return results | results_mean

    def write_members_to_file(self, lag, outdir, endswith='.nc', member=None,obsoperator=None):
        """
           :param: lag: which lag step of the filter to write, must lie in range [1,...,nlag]
           :param: outdir: directory where to write files
           :param: endswith: optional label to add to the filename, default is simply .nc
           :rtype: none

           write ensemble member information to a NetCDF file for later use. The standard output filename is
           *parameters.ddd.nc* where *ddd* is the number of the ensemble member. Standard output file location
           is the `dir.input` of the dacycle object. In principle the output file will have only two datasets inside
           called `parametervalues` which is of dimensions `nparameters` and `parametermap` which is of dimensions (180,360).
           this dataset can be read and used by a :class:`~da.baseclasses.observationoperator.ObservationOperator` object.

           .. note:: if more, or other information is needed to complete the sampling of the ObservationOperator you
                     can simply inherit from the StateVector baseclass and overwrite this write_members_to_file function.

        """


        lag=lag-1
        if member is None:
            members = self.ensemble_members[lag]
        else:
            members = [self.ensemble_members[lag][member]]

        logging.debug('write_members_to_file: dir = %s' %outdir)

        # make parallel?
        mask=[(self.dacycle['time.start']<=dt)&(dt<self.dacycle['time.end']) for dt in self.datevec]
        # logging.debug('in this cycle we will do %s' % self.datevec[mask])
        for member in members:

            filename     = os.path.join(outdir, 'parameters.%03d%s' % (member.membernumber, endswith))
            ncf          = io.CT_CDF(filename, method='create')
            dim      = ncf.add_latlon_dim()
            if self.LW:
                dim         = ncf.add_dim('time',len(self.datevec[mask])) + dim
            else:
                # "empty" timedim
                dim         = ncf.add_dim('time',1) + dim

            idx=member.membernumber
            for k,pobject in self.parameters.items():
                grid_vals=pobject.vector_to_4d(idx,self.datevec[mask],lag=lag)
                savedict              = io.std_savedict.copy()
                savedict['name']      = k 
                savedict['long_name'] = ('Statevector gridded elements for flux %s NEE values on lat-lon grid for ensemble member %s' % (k, idx))
                savedict['dims']      = dim
                savedict['values']    = grid_vals.tolist()
                savedict['units']     = 'unitless'
                ncf.add_data(savedict)

            ncf.close()

            logging.debug('successfully wrote data from ensemble member %d to file (%s) ' % (member.membernumber, filename))
        # exit()
        return


    def get_covariance(self,date=None,dacycle=None):
        """
        make a new ensemble from specified matrices, the attribute lag refers to the position in the state vector.
            note that lag=1 means an index of 0 in python, hence the notation lag-1 in the indexing below.
            the argument is thus referring to the lagged state vector as [1,2,3,4,5,..., nlag]
        """
        for k,pobject in self.parameters.items():
            pobject.compute_cov()
    
        return

    def ensemble_init(self, lag=0,cov=None):
        """
        :param lag: an integer indicating the time step in the lag order
        :param covariancematrix: a list of matrices specifying the covariance distribution to draw from
        :rtype: none

        make a new ensemble, the attribute lag refers to the position in the state vector.
        note that lag=1 means an index of 0 in python, hence the notation lag-1 in the indexing below.
        the argument is thus referring to the lagged state vector as [1,2,3,4,5,..., nlag]

        the covariance list object to be passed holds a list of matrices with a total number of dimensions [nparams, nparams], which is
        used to draw ensemblemembers from. each draw is done on a matrix from the list, to make the computational burden smaller when
        the statevector nparams becomes very large.

        """
        self.ensemble_members = list([[]]*self.nlag)
        for member in range(self.nmembers):
            new_member = EnsembleMember(member)
            # dev_matrix is initialized at zero; we add all zeros to the first member
            new_member.param_values=np.zeros((self.nparams),float,order='F')
            # logging.info("%s" % self.nparams)
            self.ensemble_members[lag].append(new_member)
        return

    def make_new_ensemble(self, lag=0,cov=None):
        """
        :param lag: an integer indicating the time step in the lag order
        :param covariancematrix: a list of matrices specifying the covariance distribution to draw from
        :rtype: none

        make a new ensemble, the attribute lag refers to the position in the state vector.
        note that lag=1 means an index of 0 in python, hence the notation lag-1 in the indexing below.
        the argument is thus referring to the lagged state vector as [1,2,3,4,5,..., nlag]

        the covariance list object to be passed holds a list of matrices with a total number of dimensions [nparams, nparams], which is
        used to draw ensemblemembers from. each draw is done on a matrix from the list, to make the computational burden smaller when
        the statevector nparams becomes very large.

        """
        # correct index
        lag=lag
        # self.ensemble_init(lag=lag)
        dof=0
        for k,pobject in self.parameters.items():
            pobject.make_new_ensemble(lag=lag)
            pobject.create_mean_statevector(lag=lag)
            dof+=pobject.dof
        
        if lag==self.nlag-1:
            self._stateparameter2members()
        # at this time the statevector doesn't contain elements. Fill it up with the default loaded_keys():
        # self._stateparameter2members(lag)
        # and set the domain_info
            self._setdomaininfo()

        logging.info('%d new ensemble members were added to the state vector # %d' % (self.nmembers, (lag + 1)))
        logging.info('statevector contains %s elements with %s degrees of freedom' % (self.nparams,dof))
        return

    def propagate(self, dacycle):
        """
        :rtype: None

        propagate the parameter values in the stateVector to the next cycle. This means a shift by one cycle
        step for all states that will
        be optimized once more, and the creation of a new ensemble for the time step that just
        comes in for the first time (step=nlag).
        in the future, this routine can incorporate a formal propagation of the statevector.

        """
        if not self.LW:
            idxlast=self.nlag-1
            # self.ensemble_members[0:idxlast-1]=self.ensemble_members[1:idxlast]
            for k,pobject in self.parameters.items():
                pobject.propagate()
            self.get_covariance()
            self.make_new_ensemble(lag=idxlast)
            # self._stateparameter2members()
            # self._setdomaininfo()
        # in lw not really progpagation
        return

    def select_parameters(self,keys):
        """
        User method to select state parameters into the statevector. Will reshape the vector.
        """
        self._select(keys)
        self._stateparameter2members()
        self._setdomaininfo()
        return

    def _select(self,keys=None):
        # Select from different parameter objects the different elements listed in keys
        if keys==None or keys=='all':
            keys=list(self.parameters_init.keys())
        self.loaded_keys=keys
        # set the indices and the load the 
        self._gather_indices()
        self.ensemble_init()
        return

    def _gather_indices(self):
        nparams_per_key=[self.parameters[key].nparams for key in self.loaded_keys ] 
        indexes=np.append([0],np.cumsum(nparams_per_key))
        self.StateParameter_indices={}
        for i,k in enumerate(self.loaded_keys):
            self.StateParameter_indices[k]=(indexes[i],indexes[i+1])
        self.nparams=np.sum(nparams_per_key)
        if not self.nlag>1:
            self.domain_info=[None]*self.nparams
        self.dacycle.dasystem['nparameters']=self.nparams
        return

    def _date2lag(self,date,qual='prior'):
        ''' Function to return the lag of the first appeareance of a statvector element. Normally 
        the last lag of the first cycle it appeared in except for at the start.
        '''
        if self.LW: return 0
        if qual=='prior':
            # get the prior value of the first appeareance 
            if advance_time(date,-(self.nlag-1 )*self.timecycle)<=self.dacycle['time.begin']:
                lag=int( (date-self.dacycle['time.begin']).days /self.timecycle)
            else:
                lag=self.nlag-1
        else: 
            lag=0
            # lag=int(nlag)-1
           
        logging.debug("found %s for day %s  at lag %s " % (qual,date,lag))
        return lag 

    def _date2datestamp(self,date,opt='prior'):
        if opt=='prior':
            # shift time
            daten=advance_time(date,-(self.nlag-1 )*self.timecycle)
            # get the prior value of the first appeareance 
        else: 
            daten=date
        b=int( (daten-self.dacycle['time.begin']).days /self.timecycle)
        d=advance_time(self.dacycle['time.begin'],b*self.timecycle)
        # case daten=date it is larger then or  equalt to d 
        # opt=1
        if daten<=self.dacycle['time.begin']:
            d=self.dacycle['time.begin']
        logging.debug("found %s for day %s file %s " % (opt,date,d ))
        return  d

    def _get_filename(self,date=None,opt='opt',shift=False):
        if self.LW:
            saved_sv = os.path.join(self.dacycle['dir.restart'], 'savestate_%s.nc' % self.dacycle['time.begin'].strftime('%Y%m%d'))
        elif date is None:
            if shift:
                saved_sv = os.path.join(self.dacycle['dir.restart'], 'savestate_%s.nc' % self.dacycle['da.restart.tstamp'].strftime('%Y%m%d'))
            else:
                saved_sv = os.path.join(self.dacycle['dir.restart'], 'savestate_%s.nc' % self.dacycle['time.start'].strftime('%Y%m%d'))
        else: 
            datestamp=self._date2datestamp(date,opt)
            saved_sv = os.path.join(self.dacycle['dir.restart'], 'savestate_%s.nc' % datestamp.strftime('%Y%m%d'))
        return saved_sv
################### end class LWstatevector ###################

if __name__ == "__main__":
    pass

