import logging
import os
chem_param_header = '''
! 
! Autogenerated by chem param parser
! JJDHooghiem
!
#define TRACEBACK write (gol,'("in ",a," (",a,", line",i5,")")') rname, __FILE__, __LINE__; call goErr
#define IF_NOTOK_RETURN(action) if (status/=0) then; TRACEBACK; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; TRACEBACK; action; return; end if
!
#include "tm5.inc"
!
!#################################################################

module chem_param

  ! --- history ----------------------------

  ! 2021-12-24 - JJDHooghiem: adapted for use with TM5-MP. Based on original work by I van der Velde.
  ! 2022-08 - AvandenBerg: added CO.

  ! --- description ------------------------

  ! definition of the chemistry: #reactions, order of species, etc.
  ! parameters needed for chemistry and rate constants

  ! --- modules ----------------------------

  use GO,    only : gol, goPr, goErr
  use binas, only : xmair
  use reaction_data
#ifndef without_photolysis
  use photolysis_data
#endif
  use deposition_data
  implicit none

integer, parameter :: nstd = 6
    real,parameter :: one=1.0
    real,parameter :: zero=0.0

  ! --- const ------------------------------

  character(len=*), parameter  ::  mname = 'chem_param'

  ! --- var --------------------------------

  ! nmark: number of 'marked' tracers
  integer, parameter :: nmark = 1
  character(len=8), dimension(nmark) :: marknam = (/ 'test    ' /)


  ! ntrace_chem: non-transported tracers
  integer,parameter :: ntrace_chem =0
    integer,parameter :: tracer_name_len=16
  ! molar mass of components
  real,    parameter :: xmh=1.0079
  real,    parameter :: xmn=14.0067

  ! Mean mass of carbon in the universe accoung for c13
  real,    parameter :: xmc=12.01115    
  real,    parameter :: xm13c=13.003355
  real,    parameter :: xm14c=14.003242

  real,    parameter :: xms=32.064
  real,    parameter :: xmo=15.9994
  real,    parameter :: xmcl=35.453

  real,    parameter :: xmco2=xmc+2*xmo 
  real,    parameter :: xm13co2=xm13c+2*xmo
  real,    parameter :: xm14co2=xm14c+2*xmo
  real,    parameter :: xmco=xmc+xmo
  real,    parameter :: xmo2=xmo+xmo
  real,    parameter :: xmn2=xmn+xmn
  real,    parameter :: xmh2o=2*xmh+xmo

  ! =============================================================
  !
  ! 2021-12-24 - JJDHooghiem: standard isotope ratios
  !                           format is as follows:
  !                           R_<standard>_<isotope>_<optioinal>
  !                           See Coplen 2002a for a compilation
  !                           As of yet, we are using approximation
  !                           explained in Tans 1993 https://doi.org/10.1029/93GB00053
  !                           Thus:
  !                           R = 13C / (13C + 12C)
  ! =============================================================
  real, parameter            :: R_VPDB_13C_TANS = 0.011112 ! Tans approximation of Craigs value see
  real, parameter            :: R_VPDB_13C = 0.01123720 ! decided to use this one

  ! =============================================================
  ! 2022-01-10 - JJDHooghiem: definitions of (approximately) constat delta values and isotope enrichments
  ! =============================================================
  real,parameter             :: delta_test = -8.
  real,parameter             :: delta_atm = -8.
  real,parameter             :: delta_ff=-30. ! global average estimate of fossil fuel carbon 13 content
  real, parameter            :: eps_ocn=-2.
  real, parameter            :: R_atm_const=(delta_atm/1000. +1. )*R_VPDB_13C
  real, parameter            :: R_ff=(delta_ff/1000. +1. )*R_VPDB_13C
  !
  ! Oxygen
  !
  real,parameter              :: ER_bio = -1.1
  real,parameter              :: ER_ff = -1.1
  real, parameter :: R_ox = 1.176e-12
  real, parameter :: N14C_atm = (0.975/(-0.008 + 1))**2
  ! used in wet_depos ...
  real, parameter            :: xmhno3=xmh + xmn + xmo*3.0


  ! ********************************************************************
  ! rates
  ! ********************************************************************
  ! heterogeneous removal rates
  ! real              :: henry(ntracet, ntemp)

  ! ********************************************************************
  ! dummy tracers
  ! ********************************************************************
  ! dummy indices for tracers not included in this chemistry;
  ! comparision of tracer index with this numbers will fail ...

  !integer, parameter  :: ico=-999
  integer, parameter  :: iso2=-999
  integer, parameter  :: iso4=-998
  integer, parameter  :: inh3=-997
  integer, parameter  :: ihno3=-996
  integer, parameter  :: ino=-995
  integer, parameter  :: ino2=-994
  integer, parameter  :: inox=-993
  integer, parameter  :: io3=-992
  integer, parameter  :: ihno4=-991
  integer, parameter  :: ih2o2=-990
  integer, parameter  :: ich4=-989

  ! ********************************************************************
  ! aerosol
  ! ********************************************************************
  real, parameter     :: density_ref=1800.0   ! for 'reference' density calculations

  ! ********************************************************************
  ! wet deposition
  ! ********************************************************************
  ! nscav       : selected species for scavenging
  ! nscav_index : index for scavenging:
  ! nscav_type  : type of scavenging:
  !               0 no scavenging
  !               1 scavenging 100 % solubility assumed
  !               2 scavenging henry solubility assumed
  !               3 scavenging, aerosol removal assumed
  !               4 scavenging, special case for SO2 with aq phase diss.
  integer, parameter  :: nscav=0
  integer, parameter  :: nscav_index(1)=(/ -999 /)
  ! note CMK CFD: wetS accounts for removal of SO2, so nscav_type = 0
  !               nh3 is taken out anyhow, since most will be scavenged
  !               by acidic falling frops below cloud
  integer, parameter  :: nscav_type(1)=(/ -999 /)
'''


chem_param_footer='''
  character(len=5),parameter :: fluxes_save='daily'
  logical,parameter :: save_emissions=.true. 
end module chem_param
'''

ftrue='.true.' 
ffalse='.false.'

masses={'co2' : 'xmco2',
        'co2c13' : 'xm13co2',
        'co2c14' : 'xm14co2',
        'n2' : 'xmn2',
        'o2' : 'xmo2',
        'co' : 'xmco'}

def l2a(ar):
    return '(/'+ar +'/)'

def consti(name, value,lines):
    lines.append('integer,parameter :: %s=%s ' % (name,value) ) 
    return
def constc(name, value,lines):
    lines.append("character(len=%s),parameter :: %s='%s' " % (len(value),name,value) ) 
    return

def farrayc(l,charlen, varname,prepend=''):
    if not isinstance(l,list):
        l=[l]
    l=["'"+ (prepend+ll).ljust(charlen)+"'" for ll in l ]
    line='character(len=%s ),dimension(%s),parameter ::  %s = %s' % (charlen, len(l), varname, l2a(','.join(l)))
    return line
def farrayi(l, varname):
    if not isinstance(l,list):
        l=[l]
    l=[str(ll) for ll in l ]
    line='integer,dimension(%s),parameter ::  %s = %s' % ( len(l), varname, l2a(','.join(l)))
    return line

def farrayr(l, varname):
    if not isinstance(l,list):
        l=[l]
    l=[str(ll) for ll in l ]
    line='real,dimension(%s),parameter ::  %s = %s' % ( len(l), varname, l2a(','.join(l)))
    return line
def farrayb(l,varname):
    if not isinstance(l,list):
        l=[l]
    for i,bo in enumerate(l):
        if bo in ['true','True','t','T']:
            l[i]=ftrue
        elif bo in ['false','False','f','F']:
            l[i]=ffalse
    line='logical,dimension(%s),parameter ::  %s = %s' % (len(l), varname, l2a(','.join(l)))
    return line

def budget2chemparam(inpd,lines):
    for key,data in inpd.items():
        logging.debug(data)
        # ems=['bg']
        ems=data['emissions'].split(' ')
        logging.debug(ems)
        lines.append(farrayc(ems,16,'names_%s' %key,prepend=key+'_'))
        lines.append(farrayc(data['from'].split(' '),24,'%s_flux_from' % key))
        lines.append(farrayb(data['d3'].split(' '),'%s_d3' % key))
        lines.append(farrayc(data['ratio'].split(' '),16,'%s_ratio' % key))
        lines.append(farrayc(data['d2d3'].split(' '),16,'%s_d2d3' % key))
    return
def getnamesmass(budgets,fwd=False,nmembers=150):
    names=[]
    xm=[]
    nmembers_species=[]
    fscale=[]
    uscale=[]
    if fwd:
        for key in budgets.keys():
            ntracers=len(budgets[key]['emissions'].split(' '))
            xm.extend([masses[key]]*ntracers)
            names.extend(['%s_%s' %(key,tr) for tr in budgets[key]['emissions'].split(' ')])
            nmembers_species.append(1)
            fscale.extend(['xmair/'+masses[key]]*ntracers)
            uscale.extend([1.0]*ntracers)
    else:
        for key in budgets.keys():
            if budgets[key]['ensemble']:
                xm.extend([masses[key]]*nmembers)
                names.extend([key]*nmembers)
                nmembers_species.append(nmembers)
                fscale.extend(['xmair/'+masses[key]]*nmembers)
                uscale.extend([1.0]*nmembers)
            else:
                xm.extend([masses[key]])
                names.extend([key])
                nmembers_species.append(1)
                fscale.extend(['xmair/'+masses[key]])
                uscale.extend([1.0])
    return names,xm,nmembers_species,fscale,uscale

def append(names,xm,key):
    return names.append(key),xm.append(masses[key])


# getnames(budgets,nmembers=150)
# for line in lines:
    # print(line)

def writechemparam(fname,settings_dict,nmembers=150,fwd=True):
    # Check if parent dir exists, create if it doesn't.
    if not os.path.exists(os.path.dirname(fname)):
        logging.error('Parent directory does not yet exist: %s. Creating it now.', os.path.dirname(fname))
        os.makedirs(os.path.dirname(fname))
        
    for key,data in settings_dict.items():
        settings_dict[key]['emissions']='bg '+data['emissions']
    lines=[]
    budget2chemparam(settings_dict,lines)
    names,xm,nmembers_species,fscale,uscale=getnamesmass(settings_dict,fwd=fwd,nmembers=nmembers)
    consti('ntracet', len(names),lines)
    consti('ntrace', len(names),lines)
    consti('maxtrace', len(names),lines)
    istd=[]
    for i,key in enumerate(settings_dict.keys()): 
        istd.append('i%s' % key)
        consti('i%s' % key, i+1,lines) 
        constc('name_%s' % key,key,lines)
    lines.append(farrayi(istd,'istd'))
    lines.append(farrayi(nmembers_species,'nmembers_species'))
    lines.append(farrayc(names,16,'names'))
    lines.append(farrayr(xm,'ra'))
    lines.append(farrayr(uscale,'uscale'))
    lines.append(farrayr(fscale,'fscale'))
    if fwd:
        constc('enkf_mode','fwd' ,lines) 
        consti('nmembers',1,lines)
    else:
        consti('nmembers',nmembers,lines)
        constc('enkf_mode','inv' ,lines)
    logging.info('Writing new file to %s' % fname)
    with open(fname,'w') as f:
        f.write(chem_param_header)
        for line in lines:
            logging.debug("Writing to chem_param.F90 : %s" % line)
            f.write('  '+line+'\n')
        f.write(chem_param_footer)
    return         




