"""CarbonTracker Data Assimilation Shell (CTDAS) Copyright (C) 2017 Wouter Peters. 
Users are recommended to contact the developers (wouter.peters@wur.nl) to receive
updates of the code. See also: http://www.carbontracker.eu. 

This program is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software Foundation, 
version 3. This program is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this 
program. If not, see <http://www.gnu.org/licenses/>."""
#!/usr/bin/env python
# tools_da.py

"""
Author : peters 

Revision History:
File created on 03 Oct 2008.

Temporary module to hold classes and methods that are in development

"""

import logging
import os
import shutil
import datetime
import re
import numpy as np

from dateutil.rrule import rrule, MO, TU, WE, TH, FR, SA, SU, YEARLY, \
     MONTHLY, WEEKLY, DAILY, HOURLY, MINUTELY, SECONDLY

HOURS_PER_DAY = 24.
MINUTES_PER_DAY  = 60.*HOURS_PER_DAY
SECONDS_PER_DAY =  60.*MINUTES_PER_DAY
MUSECONDS_PER_DAY = 1e6*SECONDS_PER_DAY
SEC_PER_MIN = 60
SEC_PER_HOUR = 3600
SEC_PER_DAY = SEC_PER_HOUR * 24
SEC_PER_WEEK = SEC_PER_DAY * 7
MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY, SUNDAY = (
    MO, TU, WE, TH, FR, SA, SU)
WEEKDAYS = (MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY, SUNDAY)

# Calendar information: days in month
DAYS_PER_MONTH = {'noleap': [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31],
                  '365_day': [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31], 
                  'standard': [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31],
                  'gregorian': [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31],
                  'proleptic_gregorian': [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31],
                  'all_leap': [0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31],
                  '366_day': [0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31],
                  '360_day': [0, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30]}


def validate_rc(rcfile, needed_items):
    """ validate the contents of an rc-file given a dictionary of required keys """

    for k, v in list(rcfile.items()):
        if v == 'True' :
            rcfile[k] = True
        if v == 'False':
            rcfile[k] = False
        if 'date' in k: 
            rcfile[k] = datetime.datetime.strptime(v, '%Y-%m-%d %H:%M:%S')

    for key in needed_items:
        if key not in rcfile:
            msg = 'Missing a required value in rc-file : %s' % key
            logging.error(msg)
            raise IOError(msg)
    logging.debug('rc-file has been validated succesfully')  


def create_dirs(dirname, forceclean=False):
    """ Create a directory and report success, only if non-existent """

    if forceclean:
        try:
            shutil.rmtree(dirname)
        except:
            pass

    # if not os.path.exists(dirname):
    # no path check here. Race conditions may occur when parallel processing is performed
    try:
        os.makedirs(dirname)
        logging.info('Creating new directory %s' % dirname)
    except FileExistsError:
        logging.debug('Using existing directory %s' % dirname)
    return dirname


def advance_time(time_in, interval,nsteps=1):
    """ Advance time_in by a specified interval"""

    time_out = time_in

    if interval in ['month','monthly']:                       # if monthly, this run will go to the first day of the next month
        nmonths = nsteps % 12 
        nyears  = int(nsteps/12)    
        if time_in.month+nmonths > 12: 
            nyears+=1      
            nmonths-=12
            # nmonths+=-1*(12-nmonths-12-time_in.month)
            # time_out = datetime.datetime(time_in.year, time_in.month + 1, 1, time_in.hour, 0, 0)
        time_out = datetime.datetime(time_in.year + nyears, time_in.month+nmonths, time_in.day, time_in.hour, 0, 0)  # end of year provision
    elif interval in [ 'year' , 'yearly']:                       # if yearly will run until next year 
        time_out = datetime.datetime(time_in.year+nsteps, time_in.month, time_in.day, time_in.hour, 0, 0)
    elif interval in [ 'week' ,'weekly']:
        time_out = time_in + datetime.timedelta(days=7)
    elif interval in [ 'day' ,'daily']:
        time_out = time_in + datetime.timedelta(days=1)
    elif isinstance(interval, datetime.timedelta):
        time_out = time_in + interval
    else:                    # assume that the interval specified is the number of days to run forward before resubmitting
        time_out = time_in + datetime.timedelta(days=float(interval))

    return time_out


def to_datetime(datestring, fmt=None):
    """ convert a date string to a datetime object """
  
    if fmt == 'TM5':
        datestring = '%04s-%02s-%02s %02s:%02s:00' % (datestring[0:4], datestring[4:6], datestring[6:8], datestring[8:10], datestring[10:12])
    elif fmt == 'pycasso-TM5':
        pass # Format already compatible
    else:
        pass 

    try:
        return datetime.datetime.strptime(datestring, '%Y-%m-%d %H:%M:%S')
    except:
        date, time = datestring.split(' ')
        year, month, day = list(map(int, date.split('-')))
        hour, minute, second = list(map(int, time.split(':')))
        return datetime.datetime(year, month, day, hour, minute, second)


def name_convert(name=None, to=None):
    """ Convert between old GLOBALVIEW-style and new ObsPack-style

      print name_convert(name="lef_surface-pfp_1", to='GV' )
            lef_01P0

      print name_convert(name="hun_35C3", to='OP' )
            hun_tower-insitu_35
   """

    identifier = 'name_convert'

    if name == None or to == None: 
        return ""
    
    platform_dict = { 'surface':0, 'shipboard':1, 'aircraft':2, 'tower':3 }
    strategy_dict = { 'flask':'D', 'pfp':'P', 'insitu':'C' }

    #----------------------------------->>>
    if to.upper() == "GV":
        # alt_surface-flask_1 -> alt_01D0
        try:
            fields = name.split('_')
            code = fields[0]
            lab_num = int(fields[2])
        except:
            return ""

        try:
            fields = fields[1].split('-')
            platform = fields[0]
            strategy = fields[1]
        except:
            return ""

        platform_num = [ v for k, v in list(platform_dict.items()) if platform.lower() == k ]
        if len(platform_num) == 0:
            print(('%s: Platform %s not found in platform dict.' % (identifier, platform)))
            return ""

        strategy_char = [ v for k, v in list(strategy_dict.items()) if strategy.lower() == k ]
        if len(strategy_char) == 0:
            print(('%s: Strategy %s not found in strategy dict.' % (identifier, strategy)))
            return "" 
        return "%s_%2.2d%1s%1d" % (code, lab_num, strategy_char[0].upper(), int(platform_num[0]))

    #----------------------------------->>>
    if to.upper() == "OP":

        # hun_35C3 -> hun_tower-insitu_35
        try:
            fields = name.split('_')
  
            code = fields[0]
            lab_num = int(fields[1][:2])
            strategy_char = fields[1][2]
            platform_num = int(fields[1][3])
        except:
            return ""

        platform = [ k for k, v in list(platform_dict.items()) if v == platform_num ]
        if len(platform) == 0:
            print(('%s: Platform number %s not found in platform dict.' % (identifier, platform_num)))
            return ""
        
        pattern = re.compile(strategy_char, re.IGNORECASE)
        strategy = [ k for k, v in list(strategy_dict.items()) if pattern.search(v) ]
        if len(strategy) == 0:
            print(('%s: Strategy character %s not found in strategy list.' % (identifier, strategy_char)))
            return ""
        return "%s_%s-%s_%d" % (code, platform[0], strategy[0], lab_num)

def _to_ordinalf(dt):
    """
    Convert :mod:`datetime` to the Gregorian date as UTC float days,
    preserving hours, minutes, seconds and microseconds.  Return value
    is a :func:`float`.
    """

    if hasattr(dt, 'tzinfo') and dt.tzinfo is not None:
        delta = dt.tzinfo.utcoffset(dt)
        if delta is not None:
            dt -= delta

    base =  float(dt.toordinal())
    if hasattr(dt, 'hour'):
        base += (dt.hour/HOURS_PER_DAY + dt.minute/MINUTES_PER_DAY +
                 dt.second/SECONDS_PER_DAY + dt.microsecond/MUSECONDS_PER_DAY
                 )
    return base

def _from_ordinalf(x, tz=None):
    """
    Convert Gregorian float of the date, preserving hours, minutes,
    seconds and microseconds.  Return value is a :class:`datetime`.
    """
    if tz is None: tz = _get_rc_timezone()
    ix = int(x)
    dt = datetime.datetime.fromordinal(ix)
    remainder = float(x) - ix
    hour, remainder = divmod(24*remainder, 1)
    minute, remainder = divmod(60*remainder, 1)
    second, remainder = divmod(60*remainder, 1)
    microsecond = int(1e6*remainder)
    if microsecond<10: microsecond=0 # compensate for rounding errors
    dt = datetime.datetime(
        dt.year, dt.month, dt.day, int(hour), int(minute), int(second),
        microsecond, tzinfo=UTC).astimezone(tz)

    if microsecond>999990:  # compensate for rounding errors
        dt += datetime.timedelta(microseconds=1e6-microsecond)

    return dt

def date2num(d,refyear=None):
    """
    *d* is either a :class:`datetime` instance or a sequence of datetimes.

    Return value is a floating point number (or sequence of floats)
    which gives the number of days (fraction part represents hours,
    minutes, seconds) since 0001-01-01 00:00:00 UTC, *plus* *one*.
    The addition of one here is a historical artifact.  Also, note
    that the Gregorian calendar is assumed; this is not universal
    practice.  For details, see the module docstring.
    """
    refyearnum=0
    if not refyear is None:
        refyearnum=_to_ordinalf(refyear)
    try: 
        return np.asarray([_to_ordinalf(val) for val in d])-refyearnum
    except:
        return _to_ordinalf(d)-refyearnum


def num2date(x, tz=None):
    """
    *x* is a float value which gives the number of days
    (fraction part represents hours, minutes, seconds) since
    0001-01-01 00:00:00 UTC *plus* *one*.
    The addition of one here is a historical artifact.  Also, note
    that the Gregorian calendar is assumed; this is not universal
    practice.  For details, see the module docstring.

    Return value is a :class:`datetime` instance in timezone *tz* (default to
    rcparams TZ value).

    If *x* is a sequence, a sequence of :class:`datetime` objects will
    be returned.
    """
    if tz is None: tz = _get_rc_timezone()
    try: 
        return [_from_ordinalf(val, tz) for val in x]
    except:
        return _from_ordinalf(x, tz)


def get_dpm(date, calendar='standard'):
    """
    Returns the number of days in the current month for a date(datetime) input objects
    """
    ndays = DAYS_PER_MONTH[calendar][date.month]
    if leap_year(date.year, calendar=calendar) and date.month == 2:
            ndays += 1
    return ndays


def leap_year(year, calendar='standard'):
    """
    Determine if year is a leap year
    """
    leap = False
    if ((calendar in ['standard', 'gregorian','proleptic_gregorian', 'julian']) and (year % 4 == 0)):
        leap = True
        if ((calendar == 'proleptic_gregorian') and (year % 100 == 0) and (year % 400 != 0)):
            leap = False
        elif ((calendar in ['standard', 'gregorian']) and (year % 100 == 0) and (year % 400 != 0)
            and (year < 1583)):
            leap = False
    return leap

def datevector( startdate, enddate,resolution='day'):
    """creates time vector with monthly interval from resolution ranging from startdate to enddate."""

    datevec = []

    # if resolution in ['year','month','monthly','yearly','week','weekly']:
    date =datetime.datetime(startdate.year,startdate.month,startdate.day)
    while date <= enddate:
        datevec.append(date)
        date=advance_time(date,resolution)

    # elif resolution in ['day','daily']:
    #     date = startdate 
    #     ndays   = (enddate-startdate).days
    #     datevec=[startdate+datetime.timedelta(days=day) for day in range(ndays)]
        # logging.debug('datevec = %s' % datevec)
    return datevec


def calc_area(lats, lons):
    """ Calculates the surface area [m2]of a grid spanned by lon and lat.
    Assumes lon,lat are evenly spaced (in degrees) vectors.
    Assumes lon,lat indicate center points
    Written by M Molen"""
    R            = 6370e3 # [m] the radius of the earth and h is the perpendicular distance from the plane containing the line of latitude to the pole.
    # grid spacing (degrees)
    dlon = np.gradient(lons).mean()
    dlat = np.gradient(lats).mean()

    # initialize A (surface area of grid point)
    nlon = len(lons)
    nlat = len(lats)
    A = np.zeros((nlat, nlon))
    for iy in range(nlat):
        Alat     = 2*np.pi*np.square(R) * np.abs( np.sin(2*np.pi*(lats[iy]-dlat/2.)/360.) -np.sin(2*np.pi*(lats[iy]+dlat/2.)/360.)) * np.abs(dlon)/360. # m2
        A[iy, :] = Alat
    return A # m2

def average_2d(arr, new_shape, 
               startlat=-90, endlat=90,
               startlon=-180, endlon=180):
    arr = np.array(arr)
    newlats = np.linspace(startlat, endlat, new_shape[-2])
    newlons = np.linspace(startlon, endlon, new_shape[-1])
    # Calculate the lats and lons of the input shape
    lats_inputshape = np.linspace(startlat, endlat, arr.shape[-2])
    lons_inputshape = np.linspace(startlon, endlon, arr.shape[-1])
    # Get the shape of the axis of which to average to
    shape = (new_shape[0], arr.shape[-2] // new_shape[-2],
             new_shape[1], arr.shape[-1] // new_shape[-1])
    # Get the area of the input lats and lons.
    new_area = calc_area(newlats, newlons)
    area = calc_area(lats_inputshape, lons_inputshape)

    # If time is included: add the time dimension to the shape
    if len(arr.shape) == 3:
        shape = (len(arr),) + shape
    # From flux/m2 to flux/gridcell
    arr = arr * area
    # Add the fluxes to the new shape
    newarr = np.nansum(arr.reshape(shape), axis=-1).sum(axis=-2)
    # And rewrite to flux/m2
    newarr = newarr/ new_area
    return newarr



if __name__ == "__main__":
    a=(datevector( datetime.datetime(2001,1,1), datetime.datetime(2002,1,1)  ,'month'))
    print(a)
    pass
